/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.flink.agents.runtime.python.operator;

import org.apache.flink.agents.api.Event;
import org.apache.flink.agents.plan.PythonFunction;
import org.apache.flink.agents.plan.actions.Action;
import org.apache.flink.agents.runtime.operator.ActionTask;
import org.apache.flink.agents.runtime.python.context.PythonRunnerContextImpl;
import org.apache.flink.agents.runtime.python.event.PythonEvent;
import org.apache.flink.agents.runtime.python.utils.PythonActionExecutor;

import static org.apache.flink.util.Preconditions.checkState;

/**
 * A special {@link ActionTask} designed to execute a Python action task.
 *
 * <p>During asynchronous execution in Python, the {@link PythonActionTask} can produce a {@link
 * PythonGeneratorActionTask} to represent the subsequent code block when needed.
 */
public class PythonActionTask extends ActionTask {

    public PythonActionTask(Object key, Event event, Action action) {
        super(key, event, action);
        checkState(action.getExec() instanceof PythonFunction);
        checkState(
                event instanceof PythonEvent,
                "Python action only accept python event, but got " + event);
    }

    public ActionTaskResult invoke(ClassLoader userCodeClassLoader, PythonActionExecutor executor)
            throws Exception {
        LOG.debug(
                "Try execute python action {} for event {} with key {}.",
                action.getName(),
                event,
                key);
        runnerContext.checkNoPendingEvents();

        String pythonAwaitableRef =
                executor.executePythonFunction(
                        (PythonFunction) action.getExec(), (PythonEvent) event, key.hashCode());
        // If a user-defined action uses an interface to submit asynchronous tasks, it will return a
        // Python coroutine (awaitable) object instance upon its first execution. Otherwise, it
        // means that no asynchronous tasks were submitted and the action has already completed.
        if (pythonAwaitableRef != null) {
            // The Python action generates an awaitable. We need to execute it once, which will
            // submit an asynchronous task and return whether the action has been completed.
            ((PythonRunnerContextImpl) runnerContext).setPythonAwaitableRef(pythonAwaitableRef);
            ActionTask tempGeneratedActionTask = new PythonGeneratorActionTask(key, event, action);
            tempGeneratedActionTask.setRunnerContext(runnerContext);
            return tempGeneratedActionTask.invoke(userCodeClassLoader, executor);
        }
        return new ActionTaskResult(
                true, runnerContext.drainEvents(event.getSourceTimestamp()), null);
    }
}
